from sys import stdout
import inspect
import traceback
import re

from time import time, sleep
class tty_colors_cmds:
  # Reset
  Color_Off='\033[0m'       # Text Reset
  
  # Regular Colors
  Black='\033[0;30m'        # Black
  Red='\033[0;31m'          # Red
  Green='\033[0;32m'        # Green
  Yellow='\033[0;33m'       # Yellow
  Blue='\033[0;34m'         # Blue
  Purple='\033[0;35m'       # Purple
  Cyan='\033[0;36m'         # Cyan
  White='\033[0;37m'        # White
  
  # Bold
  BBlack='\033[1;30m'       # Black
  BRed='\033[1;31m'         # Red
  BGreen='\033[1;32m'       # Green
  BYellow='\033[1;33m'      # Yellow
  BBlue='\033[1;34m'        # Blue
  BPurple='\033[1;35m'      # Purple
  BCyan='\033[1;36m'        # Cyan
  BWhite='\033[1;37m'       # White
  
  # Underline
  UBlack='\033[4;30m'       # Black
  URed='\033[4;31m'         # Red
  UGreen='\033[4;32m'       # Green
  UYellow='\033[4;33m'      # Yellow
  UBlue='\033[4;34m'        # Blue
  UPurple='\033[4;35m'      # Purple
  UCyan='\033[4;36m'        # Cyan
  UWhite='\033[4;37m'       # White
  
  # Background
  On_Black='\033[40m'       # Black
  On_Red='\033[41m'         # Red
  On_Green='\033[42m'       # Green
  On_Yellow='\033[43m'      # Yellow
  On_Blue='\033[44m'        # Blue
  On_Purple='\033[45m'      # Purple
  On_Cyan='\033[46m'        # Cyan
  On_White='\033[47m'       # White
  
  # High Intensity
  IBlack='\033[0;90m'       # Black
  IRed='\033[0;91m'         # Red
  IGreen='\033[0;92m'       # Green
  IYellow='\033[0;93m'      # Yellow
  IBlue='\033[0;94m'        # Blue
  IPurple='\033[0;95m'      # Purple
  ICyan='\033[0;96m'        # Cyan
  IWhite='\033[0;97m'       # White
  
  # Bold High Intensity
  BIBlack='\033[1;90m'      # Black
  BIRed='\033[1;91m'        # Red
  BIGreen='\033[1;92m'      # Green
  BIYellow='\033[1;93m'     # Yellow
  BIBlue='\033[1;94m'       # Blue
  BIPurple='\033[1;95m'     # Purple
  BICyan='\033[1;96m'       # Cyan
  BIWhite='\033[1;97m'      # White
  
  # High Intensity backgrounds
  On_IBlack='\033[0;100m'   # Black
  On_IRed='\033[0;101m'     # Red
  On_IGreen='\033[0;102m'   # Green
  On_IYellow='\033[0;103m'  # Yellow
  On_IBlue='\033[0;104m'    # Blue
  On_IPurple='\033[0;105m'  # Purple
  On_ICyan='\033[0;106m'    # Cyan
  On_IWhite='\033[0;107m'   # White


class tty_colors:
  # Reset
  Color_Off = '0'  # Text Reset
  
  # Regular Colors
  Black = '30'  # Black
  Red = '31'  # Red
  Green = '32'  # Green
  Yellow = '33'  # Yellow
  Blue = '34'  # Blue
  Purple = '35'  # Purple
  Cyan = '36'  # Cyan
  White = '37'  # White
  
  # High Intensity
  IBlack = '90'  # Black
  IRed = '91'  # Red
  IGreen = '92'  # Green
  IYellow = '93'  # Yellow
  IBlue = '94'  # Blue
  IPurple = '95'  # Purple
  ICyan = '96'  # Cyan
  IWhite = '97'  # White
  
  # Background
  On_Black = '40'  # Black
  On_Red = '41'  # Red
  On_Green = '42'  # Green
  On_Yellow = '43'  # Yellow
  On_Blue = '44'  # Blue
  On_Purple = '45'  # Purple
  On_Cyan = '46'  # Cyan
  On_White = '47'  # White
  
  # High Intensity backgrounds
  On_IBlack = '100'  # Black
  On_IRed = '101'  # Red
  On_IGreen = '102'  # Green
  On_IYellow = '103'  # Yellow
  On_IBlue = '104'  # Blue
  On_IPurple = '105'  # Purple
  On_ICyan = '106'  # Cyan
  On_IWhite = '107'  # White


def print_color(txt, color=tty_colors_cmds.Color_Off):
  out = '%s%s'%(color, txt)
  stdout.write(out)
  #reset
  stdout.write(tty_colors_cmds.Color_Off)
  stdout.write('\n')

def write_color(txt, color=tty_colors_cmds.Color_Off):
  out = '%s%s'%(color, txt)
  stdout.write(out)
  #reset
  stdout.write(tty_colors_cmds.Color_Off)

def get_colored(txt, fg_color='', bg_color='', Bold=False):
  B = '1' if Bold else '0'
  fg = fg_color if fg_color == '' else ';' + fg_color
  bg = bg_color if bg_color == '' else ';' + bg_color
  if type(txt) is unicode:
    txt = txt.encode('UTF-8')
  ttycmd = '\033[{B}{fg}{bg}m{txt}\033[0m'.format(B=B,fg=fg,bg=bg, txt=txt)
  return ttycmd
  
def tty_reset():
  stdout.write('\e[0m')

def get_regx_spans(txt, regxpattern):
  spans = []
  iterV = regxpattern.finditer(txt)
  for m in iterV.next():
    span = m.span()
    spans.append(span)
  if len(spans)> 0:
    return spans
  else:
    return None

def Highlight_custom(txt, pattern, color):
  #type:(str, re._pattern_type, tuple[str]) -> str
  if type(txt) is unicode:
    txt = txt.encode('UTF-8')
  retV = txt
  if type(pattern) is re._pattern_type:
    founds = pattern.findall(txt)
    newtxt = txt
    for s in founds:
      colored_s = get_colored(s, color[0], color[1], Bold=True)
      newtxt = newtxt.replace(s, colored_s)
    retV = newtxt
  elif type(pattern) is str:
    s = pattern
    colored_s = get_colored(s, color[0], color[1], Bold=True)
    newtxt = txt.replace(s, colored_s)
    retV = newtxt

  return retV

  # vars = re.findall(r'\$\((\w+)\)', txt)
  # newtxt = re.sub(r'\$\((\w+)\)', r'{\1}', txt)
  #
  # repWith = get_colored('Warning', tty_colors.Yellow, tty_colors.On_IBlack, Bold=True)
  # reTV = txt.replace('warning', repWith)
  # reTV = reTV.replace('Warning', repWith)
  # return reTV

def HighlightWarnings(txt):
  #type:(str) -> str

  repWith = get_colored('Warning', tty_colors.Yellow, tty_colors.On_IBlack, Bold=True)
  reTV = txt.replace('warning', repWith)
  reTV = reTV.replace('Warning', repWith)
  return reTV

def HighlightErrors(txt):
  #type:(str) -> str
  repWith = get_colored('Error', tty_colors.Red, tty_colors.On_IBlack, Bold=True)
  reTV = txt.replace('error', repWith)
  reTV = reTV.replace('Error', repWith)
  return reTV

def HighlightNotes(txt):
  #type:(str) -> str
  repWith = get_colored('Note', tty_colors.Cyan, Bold=True)
  reTV = txt.replace('note', repWith)
  reTV = reTV.replace('Note', repWith)

  repWith = get_colored('Info', tty_colors.Cyan, Bold=True)
  reTV = reTV.replace('info', repWith)
  reTV = reTV.replace('Info', repWith)

  return reTV

def get_makefile_var(var_str):
  """
  :param var_str: str
  :return:
  """
  outerframe = inspect.stack()[1][0]
  outerframe = outerframe.f_back.f_back
  outerframeGlobals = outerframe.f_globals

  try:
    var = outerframeGlobals[var_str]
    return var
  except:
    return None

def Print_Debuging_messages():
  print_color('Debugging message: (Program Exception)', tty_colors_cmds.On_Red)
  traceback.print_exc()
  
def is_Highlight_ON():
  outerframe = inspect.stack()[1][0]
  preFrame = outerframe.f_globals
  outerframe = outerframe.f_back
  outerframeGlobals = outerframe.f_globals

  try:
    Highlighting = preFrame['_Highlighting']
    if Highlighting:
      return True
  except:
    pass

  try:
    HighlightWarnings = outerframeGlobals['HighlightWarnings']
    if HighlightWarnings:
      return True
  except:
    pass
  
  try:
    HighlightErrors = outerframeGlobals['HighlightErrors']
    if HighlightErrors:
      return True
  except:
    pass
  
  try:
    HighlightNotes = outerframeGlobals['HighlightNotes']
    if HighlightNotes:
      return True
  except:
    pass

  return False

def wait_process(Timeout, Proc, Print_statusTime = -1):
  cindex = 0
  timeStep = 0.1
  sec_count = 1.0
  CountDown = Timeout
  
  T1 = time()
  tdiff = time() - T1
  while tdiff < Timeout:
    alive = Proc.poll()
    if alive is not None:
      stdout.write('\r                          \r')
      return False
    sleep(timeStep); sec_count -= timeStep
    tdiff = time() - T1
    CountDown = int(Timeout - tdiff)
    if tdiff >= Print_statusTime and sec_count <=0:
      sec_count = 1.0
      if cindex == 0:
        stdout.write('\r                          \r')
        write_color('\rwaiting... [%d]'%CountDown, tty_colors_cmds.On_Yellow); stdout.flush()
        cindex = 1
      else:
        stdout.write('\r                          \r')
        write_color('\rwaiting... [%d]'%CountDown, tty_colors_cmds.On_Cyan) ; stdout.flush()
        cindex = 0
      

  stdout.write('\r          \r')
  return True # The Process is Timed Out

def kill_alive_process(Proc):
  alive = Proc.poll()
  if alive is None:
    try:
      Proc.kill()
    except:
      pass
  


#from distutils.core import setup
import os
from setuptools import setup
from setuptools.command.install import install

# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

class MyInstall(install):
    def run(self):
        install.run(self)
        #print("\n\n\n\nI did it!!!!\n\n\n\n")
        cmd = 'sudo activate-global-python-argcomplete'
        print cmd
        os.system(cmd)        

setup(
    name="pymake2",
    version = "0.5.33",
    author = "Saud Wasly",
    author_email = "saudalwasli@gmail.com",
    description = ("pymake2 is a simple Python-based make system. It brings simplicity and flexibility of Python language to makefiles."),
    license = "MIT",
    keywords = "make makefile build",
    url = "https://bitbucket.org/saudalwasly/pymake2/src",
    install_requires=["argcomplete", "sarge"],
    packages=['pymake2'],
    scripts=['pymake2/pymake2', 'pymake2/pmake2'],
    #data_files = [('', ['__init__.py', 'pymake2', 'make.py', 'utility.py', 'makefile_template.py'])] ,
    long_description=read('README.rst'),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Topic :: Utilities",        
        "License :: OSI Approved :: MIT License",
    ],
    cmdclass={'install': MyInstall}    
)


__author__ = 'Saud Wasly'

import sys
import os
import sarge
import re
import inspect
import utility as util
from utility import tty_colors as colors
import fnmatch

from time import sleep
# makefileM = None # to be assigned upon importing

_Highlighting = False
_HighlightingDict = {}


def eval(txt):
  outerframe = inspect.stack()[1][0]
  outerframeGlobals = outerframe.f_globals
  
  vars = re.findall(r'\$\((\w+)\)', txt)
  newtxt = re.sub(r'\$\((\w+)\)', r'{\1}', txt)
  
  for v in vars:
    try:
      val = outerframeGlobals[v]
      if type(val) is list:
        val = ' '.join(val)
      newtxt = newtxt.replace('{%s}'%v, val)
    except:
      val = os.getenv(v)
      if val:
        newtxt = newtxt.replace('{%s}' % v, val)
      else:
        util.print_color('Error: cannot find variable %s' % v, util.tty_colors_cmds.Red)
        sys.exit()
  return newtxt

def printcolor(txt, fg='', bg='', B=False):
  msg = util.get_colored(txt, fg, bg, B)
  print msg

def regx(pattern):
  retV = re.compile(pattern, flags=re.IGNORECASE)
  return retV

def hl(regxP, fg_color=colors.Yellow, bg_color = ''):
  global _Highlighting, _HighlightingDict
  _Highlighting = True
  _HighlightingDict[regxP] = (fg_color, bg_color)


def _Highlight_Outputs(txt):
  outerframe = inspect.stack()[1][0]
  outerframe = outerframe.f_back
  outerframeGlobals = outerframe.f_globals

  retV = txt
  try:
    HighlightWarnings = outerframeGlobals['HighlightWarnings']
    if HighlightWarnings:
      retV = util.HighlightWarnings(retV)
  except:
    pass

  try:
    HighlightErrors = outerframeGlobals['HighlightErrors']
    if HighlightErrors:
      retV = util.HighlightErrors(retV)
  except:
    pass

  try:
    HighlightNotes = outerframeGlobals['HighlightNotes']
    if HighlightNotes:
      retV = util.HighlightNotes(retV)
  except:
    pass


  try: #custom highlight

    if _Highlighting:
      for key in _HighlightingDict.keys():
        color = _HighlightingDict[key]
        retV = util.Highlight_custom(retV, key, color)
  except Exception as e:
    print e
    pass

  return retV

def find(root='./', filter='*', recursive = False, abslute = False, DirOnly = False):
  srcfiles = []
  rootdir = os.path.abspath(root) if abslute else os.path.normpath(root)
  if recursive:
    for cur_root, dir, files in os.walk(rootdir):
      if DirOnly:
        srcfiles.extend([cur_root + '/' + e for e in dir])
      else:
        for srcfile in fnmatch.filter(files, filter):
          srcfiles.append(cur_root + '/' + srcfile)
  else:
    for cur_root, dir, files in os.walk(rootdir):
      if DirOnly:
        srcfiles.extend([cur_root + '/' + e for e in dir])
      else:
        for srcfile in fnmatch.filter(files, filter):
          srcfiles.append(cur_root + '/' + srcfile)
      break

  if DirOnly:
    # add the current root directory to the list
    rootdir = os.path.abspath(root) if abslute else os.path.normpath(root)
    srcfiles.append(rootdir)
  return srcfiles


def get_dir(path):
  if type(path) is str:
    dir_str = os.path.dirname(path)
    return dir_str
  elif type(path) is list:
    retPaths = []
    for p in path:
      dir_str = os.path.dirname(p)
      retPaths.append(dir_str)
    return retPaths
  else:
    return None
  
def printlist(lst):
  if type(lst) is not list:
    lst = lst.split()
  for i in lst:
    print i
  
  return lst
    
def get_filename(path):
  if type(path) is str:
    dir_str = os.path.basename(path)
    return dir_str
  elif type(path) is list:
    retPaths = []
    for p in path:
      dir_str = os.path.basename(p)
      retPaths.append(dir_str)
    return retPaths
  else:
    return None
   
  
def compile(compiler, flags, sources, objects):
  # utl.print_color('Compiling ...', utl.tty_colors.On_Cyan)
  Highlight_NO = True if util.is_Highlight_ON() else False
    
  if type(sources) is list:
    srcs = sources
  else: # in the case of str
    srcs = sources.split()
  
  if type(objects) is list:
    objs = objects
  else: # in the case of str
    objs= objects.split()

  if len(srcs) != len(objs):
    util.write_color('Error: ', util.tty_colors_cmds.Red)
    print 'the length of the source files list does not match with objects files list'
    return
  
  for i, item in enumerate(srcs):
    cmd = '{CC} {flags} -c {src} -o {obj}'.format(CC=compiler, flags=flags, src=item, obj=objs[i])
    
    srcFile = os.path.basename(item)
    objFile = os.path.basename(objs[i])
    srcFile = srcFile.split('.')[0]
    objFile = objFile.split('.')[0]
    if srcFile != objFile:
      util.write_color('Compiling Error: ', util.tty_colors_cmds.BRed)
      print 'source file %s and object file %s do not match. Make sure that the source and the object files lists are correspondent'%(item, objs[i])
      return False
    if os.path.isfile(objs[i]): # if the object file already exists
      src_mTime = os.path.getmtime(item)
      obj_mtime = os.path.getmtime(objs[i])
      if src_mTime <= obj_mtime:
        continue
    else: # no obj file exists
        objDir = os.path.dirname(objs[i])
        objDir = os.path.normpath(objDir)
        if not os.path.exists(objDir):
          os.makedirs(objDir)
   
    util.print_color('Compiling: %s' % item, util.tty_colors_cmds.On_Cyan)
    success, outputs = sh(cmd, True, Highlight_NO)
    if Highlight_NO:
      print(_Highlight_Outputs(outputs))
      
    if not success:
    # if not run(cmd, show_cmd=True):
      util.write_color('Error: ', util.tty_colors_cmds.BRed)
      print 'failed to compile, \n  %s'%cmd
      return False
  
  return True
    
def link(linker, flags, objects, executable):
  if type(objects) is list:
    objs = ' '.join(objects)
  else:
    objs = objects.strip().strip('\n')
    
  objectsList = objs.split()

  linkFlag = True
  if os.path.isfile(executable):
    linkFlag = False
    exe_mTime = os.path.getmtime(executable)
    for obj in objectsList:
      obj_mtime = os.path.getmtime(obj)
      if obj_mtime > exe_mTime:
        linkFlag = True
        break
  
  if linkFlag:
    util.print_color('Linking ...', util.tty_colors_cmds.On_Blue)
    cmd = '{linker} {flags} {objs} -o {executable}'.format(linker=linker, flags=flags, objs=objs, executable=executable)
    hl = util.is_Highlight_ON()
    success, outputs = sh(cmd, True, hl)
    if hl:
      print(_Highlight_Outputs(outputs))
    
    if not success:
      util.print_color("Failed to link object files to assemble '%s'"%executable, util.tty_colors_cmds.BRed)
      return False
    else:
      return True
  else:
    return True
  
def archive(archiver, flags, objects, library):
  if type(objects) is list:
    objs = ' '.join(objects)
  else:
    objs = objects
    
  objectsList = objs.split()

  satisfactionFlag = False
  if os.path.isfile(library):
    satisfactionFlag = True
    output_mTime = os.path.getmtime(library)
    for obj in objectsList:
      obj_mtime = os.path.getmtime(obj)
      if obj_mtime > output_mTime:
        satisfactionFlag = False
        break
  
  if not satisfactionFlag:
    util.print_color('Archiving...', util.tty_colors_cmds.On_Blue)
    cmd = '{AR} {flags} {output} {objs}'.format(AR=archiver, flags=flags, objs=objs, output=library)
    hl = util.is_Highlight_ON()
    success, outputs = sh(cmd, True, hl)
    if hl:
      print(_Highlight_Outputs(outputs))
    
    if not success:
      util.print_color("Failed to archive object files to assemble '%s'" % library, util.tty_colors_cmds.BRed)
      return False
    else:
      return True
  else:
    return True
    

def normpaths(paths):
  if type(paths) is str:
    dir_str = os.path.normpath(paths)
    return dir_str
  elif type(paths) is list:
    retPaths = []
    for p in paths:
      dir_str = os.path.normpath(p)
      retPaths.append(dir_str)
    return retPaths
  else:
    return None

def join(*args):
    try:
      retV = ' '.join(args) # this works if all args are str type
      return retV
    except: # deal with different types
      retV = ''
      for arg in args:
        if type(arg) is list:
          argItems = ' '.join(arg)
          retV += argItems + ' '
            
        else: # assume str
          retV += arg + ' '
      return retV
      
def replace(srclist, term, repwith):
  if type(srclist) is list:
    retV = []
    for item in srclist:
      x = item.replace(term, repwith)
      retV.append(x)
    
    return retV
  else:
    retV = srclist.replace(term, repwith)
    return retV

def retarget(srclist, targetP, omit=''):
  if targetP.endswith('/'):
    targetP = targetP[:-1]

  if type(srclist) is list:
    retV = []
    for item in srclist:
      x = item.replace(omit, '')
      x = targetP + '/' + x
      x = os.path.normpath(x)
      retV.append(x)

    return retV
  else:
    x = srclist.replace(omit, '')
    x = targetP + '/' + x
    x = os.path.normpath(x)
    retV = x
    return retV

def exclude(original, ignors):
  retV = []
  for item in original:
    if item not in ignors:
      retV.append(item)
  return retV
      
def shell(cmd):
  P = sarge.run(cmd, shell=True, stdout=sarge.Capture())
  return P.stdout.text

def sh(cmd, show_cmd=False, CaptureOutput = False, Timeout = -1):
  if show_cmd:
    print(cmd)
  try:
    if CaptureOutput:
      if Timeout > -1:
        P = sarge.run(cmd, shell=True, stdout=sarge.Capture(), stderr=sarge.Capture(), async=True)
        # sleep(3)
        try:
          CMD = P.commands[0] #type: sarge.Command # FIXME: This line generates index exception sometime
          timed_out = util.wait_process(Timeout, CMD)
          if timed_out:
            util.print_color('The command "%s" is timed out!'%cmd, util.tty_colors_cmds.On_Red)
          util.kill_alive_process(CMD)
        except:
          pass
      else:
        P = sarge.run(cmd, shell=True, stdout=sarge.Capture(), stderr=sarge.Capture())
    else:
      if Timeout > -1:
        P = sarge.run(cmd, shell=True, async=True)
        # sleep(3)
        try:
          CMD = P.commands[0] #type: sarge.Command # FIXME: This line generates index exception sometime
          timed_out = util.wait_process(Timeout, CMD)
          if timed_out:
            util.print_color('The command "%s" is timed out!'%cmd, util.tty_colors_cmds.On_Red)
          util.kill_alive_process(CMD)
        except:
          pass
      else:
        P = sarge.run(cmd, shell=True)
    
    outputs = ''

    if P.stdout and len(P.stdout.text) > 0:
      outputs = P.stdout.text
    if P.stderr and len(P.stderr.text) > 0:
      if outputs == '':
        outputs = P.stderr.text
      else:
        outputs += '\n' + P.stderr.text
    return P.returncode == 0, outputs
  except:
    if util.get_makefile_var('Debug') == True:
      util.Print_Debuging_messages()
  
    return False, ''
    
  
  
def run(cmd, show_cmd=False, Highlight=False, Timeout = 10):
  """
  :param cmd: (str) the shell command
  :param show_cmd: (bool) print the command before executing it
  :param Highlight: (bool) apply color highlights for the outputs
  :param Timeout: (float) any positive number in seconds
  :return:
  """
  if Highlight:
    success, outputs = sh(cmd, show_cmd, True, Timeout)
    hl_out = _Highlight_Outputs(outputs)
    print(hl_out)
  else:
    success, outputs = sh(cmd, show_cmd, False, Timeout)
  
  return success

def target(func):
  """
    This is a decorator function
  :param func:
  :return:
  """
  def target_func(*original_args, **original_kwargs):
    # print 'before the func'
    # print original_kwargs
    retV =  func(*original_args, **original_kwargs)
    if retV is None or retV == False:
      return False
    else:
      return True
    # print 'after the func'
  
  return target_func


if __name__ == '__main__':
    print 'testing find()'
    flist = find(DirOnly=True, abslute=False)
    printlist(flist)
    print 'File List has %d items'%len(flist)